import argparse
from numbers import Number

import numpy as np
import torch
from tqdm import tqdm

from .summarize import SummarizerCollection
from .utils import print_num_params


class BaseExperimentManager:
    """
    Experiment manager.

    Data attributes:
    - 'args': argparse.Namespace containing all config parameters. When
      initializing this object, if 'args' is not given, all config
      parameters are set based on experiment defaults and user input, using
      argparse.
    - 'run_description': string description of the run that includes a timestamp
      and can be used e.g. as folder name for logging.
    - 'model': the model.
    - 'device': torch.device that is being used
    - 'dataloaders': DataLoaders, with attributes 'train' and 'test'
    - 'optimizer': the optimizer
    """

    def __init__(self, args=None):
        self.device = None
        self.dataloaders = None
        self.model = None
        self.optimizer = None
        self.args = args
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            allow_abbrev=False)
        if args is None:
            self.args = self._parse_args(parser)
        self.run_description = self._make_run_description(self.args)


    def _parse_args(self, parser):
        """
        Parse command-line arguments defining experiment settings.

        :return: args: argparse.Namespace with experiment settings
        """
        raise NotImplementedError


    def add_required_args(self,
                          parser,
                          batch_size=None,
                          test_batch_size=None,
                          lr=None,
                          max_epochs=1000000,
                          max_steps=2147483647,
                          seed=54321,
                          train_log_every=10000,
                          test_log_every=10000,
                          checkpoint_every=100000,
                          keep_checkpoint_max=3,
                          resume="",
                          ):

        parser.add_argument('--batch-size',
                            type=int,
                            default=batch_size,
                            metavar='N',
                            dest='batch_size',
                            help='training batch size')

        parser.add_argument('--test-batch-size',
                            type=int,
                            default=test_batch_size,
                            metavar='N',
                            dest='test_batch_size',
                            help='test batch size')

        parser.add_argument('--lr',
                            type=float,
                            default=lr,
                            metavar='LR',
                            dest='lr',
                            help='learning rate')

        parser.add_argument('--seed',
                            type=int,
                            default=seed,
                            metavar='N',
                            dest='seed',
                            help='random seed')

        parser.add_argument('--tr-log-every',
                            type=int,
                            default=train_log_every,
                            metavar='N',
                            dest='train_log_every',
                            help='log training metrics every this number of '
                                 'training steps')

        parser.add_argument('--ts-log-every',
                            type=int,
                            default=test_log_every,
                            metavar='N',
                            dest='test_log_every',
                            help='log test metrics every this number of '
                                 'training steps')

        parser.add_argument('--checkpoint-every',
                            type=int,
                            default=checkpoint_every,
                            metavar='N',
                            dest='checkpoint_every',
                            help='save model checkpoint every this number of '
                                 'training steps')

        parser.add_argument('--keep-checkpoint-max',
                            type=int,
                            default=keep_checkpoint_max,
                            metavar='N',
                            dest='keep_checkpoint_max',
                            help='keep at most this number of most recent '
                                 'model checkpoints')

        parser.add_argument('--max-steps',
                            type=int,
                            default=max_steps,
                            metavar='N',
                            dest='max_steps',
                            help='max number of training steps')

        parser.add_argument('--max-epochs',
                            type=int,
                            default=max_epochs,
                            metavar='N',
                            dest='max_epochs',
                            help='max number of training epochs')

        parser.add_argument('--nocuda',
                            action='store_true',
                            dest='no_cuda',
                            help='do not use cuda')

        parser.add_argument('--descr',
                            type=str,
                            default='',
                            metavar='STR',
                            dest='additional_descr',
                            help='additional description for experiment name')

        parser.add_argument('--dry-run',
                            action='store_true',
                            dest='dry_run',
                            help='do not save anything to disk')

        parser.add_argument('--resume',
                            type=str,
                            metavar='NAME',
                            default=resume,
                            dest='resume',
                            help="load the run with this name and resume training")


    @staticmethod
    def _make_run_description(args):
        """
        Create a string description of the run. It is used in the names of the
        logging folders.

        :param args: experiment config
        :return: the run description
        """
        raise NotImplementedError


    def make_datamanager(self):
        """
        Create a DatasetManager object. To be overridden.
        :return: DatasetManager
        """
        raise NotImplementedError


    def make_model(self):
        """
        Create a model. To be overridden.
        :return: model
        """
        raise NotImplementedError


    def make_optimizer(self):
        """
        Create an optimizer. To be overridden.
        :return: optimizer
        """
        raise NotImplementedError


    def setup(self, checkpoint_folder=None):
        """
        Setup experiment: load dataset, create model, load weights if a
        checkpoint is given, create optimizer.
        """

        # If checkpoint folder is given, load model from there to resume
        resume = checkpoint_folder is not None

        # Dataset
        print("Getting dataset ready...")
        self.dataloaders = self.make_datamanager()
        print("Data shape: {}".format(self.dataloaders.data_shape))
        print("Train/test set size: {}/{}".format(
            len(self.dataloaders.train.dataset),
            len(self.dataloaders.test.dataset),
        ))

        # Model
        print("Creating model...")
        self.model = self.make_model().to(self.device)
        print_num_params(self.model)

        # Load weights if resuming training
        if resume:
            self.load_model(checkpoint_folder, step=None)

        # Optimizer
        self.optimizer = self.make_optimizer()


    def load_model(self, checkpoint_folder, step=None):
        """
        Loads model weights from a checkpoint in the specified folder.
        If step is given, it attempts to load the checkpoint at that step.
        The weights are loaded with map_location=device, where device is the
        current device of this experiment.
        """
        self.model.load(checkpoint_folder, self.device, step=step)


    def forward_pass(self, x, y=None):
        """
        Simple single-pass model evaluation. It consists of a forward pass
        and computation of all necessary losses and metrics.
        """
        raise NotImplementedError


    def post_backward_callback(self):
        pass


    @classmethod
    def get_metrics_dict(cls, results):
        """
        Given a dict of results, return a dict of metrics to be given to
        summarizers. Keys are also used as names for tensorboard logging.

        In the base implementation, keys are simply copyed and used as names
        for tensorboard.

        Only scalars accepted, non-scalars are discarded. Actually, anything
        that either is a scalar or has the method item(). That should include
        Python scalars, numpy scalars, torch scalars, numpy and torch
        arrays/tensors with one element.

        Override to customize translation from results to dictionary of
        scalar metrics.
        """
        metrics_dict = {}
        for k in results:
            x = results[k]
            if isinstance(x, Number):
                metrics_dict[k] = x
                continue
            try:
                metrics_dict[k] = x.item()
            except (AttributeError, ValueError):  # is this enough?
                pass
        return metrics_dict


    @classmethod
    def train_log_str(cls, summaries, step, epoch=None):
        s = "       [step {}]".format(step)
        for k in summaries:
            s += "  {key}={value:.5g}".format(key=k, value=summaries[k])
        return s


    @classmethod
    def test_log_str(cls, summaries, step, epoch=None):
        s = "       "
        if epoch is not None:
            s += "[step {}, epoch {}]".format(step, epoch)
        for k in summaries:
            s += "  {key}={value:.5g}".format(key=k, value=summaries[k])
        return s


    def test_procedure(self, **kwargs):
        """
        Execute test procedure for the experiment. This typically includes
        collecting metrics on the test set using forward_pass().
        For example in variational inference we might be interested in
        repeating this many times to derive the importance-weighted ELBO.

        :return: summaries (dict)
        """
        raise NotImplementedError


    def additional_testing(self, img_folder):
        """
        Perform additional testing, including possibly generating images.

        :param img_folder: folder to store images
        """
        raise NotImplementedError


class VIExperimentManager(BaseExperimentManager):
    """
    Variational inference experiment manager. This version implements a default
    test_procedure for variational inference.

    Data attributes:
    - 'args': argparse.Namespace containing all config parameters. When
      initializing this object, if 'args' is not given, all config
      parameters are set based on experiment defaults and user input, using
      argparse.
    - 'run_description': string description of the run that includes a timestamp
      and can be used e.g. as folder name for logging.
    - 'model': the model.
    - 'device': torch.device that is being used
    - 'dataloaders': DataLoaders, with attributes 'train' and 'test'
    - 'optimizer': the optimizer
    """

    def test_procedure(self, iw_samples=None):
        """
        Execute test procedure for the experiment. This typically includes
        collecting metrics on the test set using forward_pass().
        Repeat this many times to derive the importance-weighted ELBO.

        :param iw_samples: number of samples for the importance-weighted ELBO.
                The other metrics are also averaged over all these samples,
                yielding a more accurate estimate.
        :return: summaries (dict)
        """

        # Shorthand
        test_loader = self.dataloaders.test
        step = self.model.global_step
        args = self.args
        n_test = len(test_loader.dataset)

        # If it's time to estimate log likelihood, use many samples.
        # If given, use the given number.
        if iw_samples is None:
            iw_samples = 1
            if step % args.loglikelihood_every == 0 and step > 0:
                iw_samples = args.loglikelihood_samples

        # Setup
        summarizers = SummarizerCollection(mode='sum')
        progress = tqdm(total=len(test_loader) * iw_samples, desc='test ')
        all_elbo_sep = torch.zeros(n_test, iw_samples)
        for batch_idx, (x, y) in enumerate(test_loader):
            for i in range(iw_samples):
                outputs = self.forward_pass(x, y)

                # elbo_sep shape (batch size,)
                i_start = batch_idx * args.test_batch_size
                i_end = (batch_idx + 1) * args.test_batch_size
                all_elbo_sep[i_start: i_end, i] = outputs['elbo_sep'].detach()

                metrics_dict = self.get_metrics_dict(outputs)
                multiplier = (x.size(0) / n_test) / iw_samples
                for k in metrics_dict:
                    metrics_dict[k] *= multiplier
                summarizers.add(metrics_dict)

                progress.update()
        progress.close()

        if iw_samples > 1:
            # Shape (test set size,)
            elbo_iw = torch.logsumexp(all_elbo_sep, dim=1)
            elbo_iw = elbo_iw - np.log(iw_samples)

            # Mean over test set (scalar)
            elbo_iw = elbo_iw.mean().item()
            key = 'elbo/elbo_IW_{}'.format(iw_samples)
            summarizers.add({key: elbo_iw})

        summaries = summarizers.get_all(reset=True)

        return summaries


    def add_required_args(self,
                          parser,
                          loglikelihood_every=50000,
                          loglikelihood_samples=100,
                          **kwargs
                          ):

        super().add_required_args(parser, **kwargs)

        parser.add_argument('--ll-every',
                            type=int,
                            default=loglikelihood_every,
                            metavar='N',
                            dest='loglikelihood_every',
                            help='evaluate log likelihood every this number '
                                 'of training steps')

        parser.add_argument('--ll-samples',
                            type=int,
                            default=loglikelihood_samples,
                            metavar='N',
                            dest='loglikelihood_samples',
                            help='number of importance-weighted samples to '
                                 'evaluate log likelihood')



import datetime
import os
import random
import re
from collections import OrderedDict

import numpy as np
import torch
import torch.utils.data

from .options import get_options

__sentinel = object()

def print_num_params(model, max_depth=__sentinel):

    if max_depth is __sentinel:
        max_depth = get_options('model_print_depth')
    assert max_depth is None or isinstance(max_depth, int)

    sep = '.'  # string separator in parameter name
    print("\n--- Trainable parameters:")
    num_params_tot = 0
    num_params_dict = OrderedDict()

    for name, param in model.named_parameters():
        if not param.requires_grad:
            continue

        num_params = param.numel()

        if max_depth is not None:
            split = name.split(sep)
            prefix = sep.join(split[:max_depth])
        else:
            prefix = name
        if prefix not in num_params_dict:
            num_params_dict[prefix] = 0
        num_params_dict[prefix] += num_params
        num_params_tot += num_params
    for n, n_par in num_params_dict.items():
        print("{:8d}  {}".format(n_par, n))
    print("  - Total trainable parameters:", num_params_tot)
    print("---------\n")


def set_rnd_seed(seed, aggressive=False):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

    # The two lines below might slow down training
    if aggressive:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_date_str():
    return datetime.datetime.now().strftime("%y%m%d_%H%M%S")


def linear_anneal(x, start, end, steps):
    assert x >= 0
    assert steps > 0
    assert start >= 0
    assert end >= 0
    if x > steps:
        return end
    if x < 0:
        return start
    return start + (end - start) / steps * x


def to_np(x):
    try:
        return x.detach().cpu().numpy()
    except AttributeError:
        return x


def get_module_device(module):
    return next(module.parameters()).device


def is_conv(m):
    return isinstance(m, torch.nn.modules.conv._ConvNd)


def is_linear(m):
    return isinstance(m, torch.nn.Linear)


def named_leaf_modules(module):
    # Should work under common naming assumptions, but it's not guaranteed
    last_name = ''
    for name, l in reversed(list(module.named_modules())):
        if name not in last_name:
            last_name = name
            yield name, l


def checkpoints_in_folder(folder):
    """
    Find checkpoint files in the speficied folder, return a list of file names
    and a list of integers corresponding to the time steps when the checkpoints
    were saved. These lists have the same length, they have a one-to-one
    correspondence, and they are sorted by number.

    Parameters
    ----------
    folder: str

    Returns
    -------
    filenames: list of str
        File names of checkpoints. The full path is given by joining the
        input argument folder with these file names.
    numbers: list of int

    """

    def is_checkpoint_file(f):
        # TODO use regex properly and maybe get number directly here
        full_path = os.path.join(folder, f)
        return (os.path.isfile(full_path) and
                f.startswith("model_") and
                f.endswith('.pt'))

    filenames = [f for f in os.listdir(folder) if is_checkpoint_file(f)]
    regex = re.compile(r'\d+')
    numbers = list([int(regex.search(n).group(0)) for n in filenames])
    assert len(filenames) == len(numbers)
    filenames = list([filenames[i] for i in np.argsort(numbers)])
    return filenames, numbers



from copy import deepcopy

_print_options = {
    'model_print_depth': 3,
}

def set_options(**kwargs):
    # Check arguments
    invalid = []
    for k in kwargs:
        if k not in _print_options:
            invalid.append(k)
    if len(invalid) > 0:
        msg = "Invalid option names: "
        invalid = [repr(k) for k in invalid]
        msg = msg + ", ".join(invalid)
        raise ValueError(msg)

    for k, v in kwargs.items():
        if not _check_option(k, v):
            msg = "Invalid value {} for option {}".format(repr(v), repr(k))
            raise ValueError(msg)

    # Update options
    _print_options.update(**kwargs)

def get_options(name=None):
    if name is None:
        return deepcopy(_print_options)
    return deepcopy(_print_options[name])

def _check_option(name, value):
    if name == 'model_print_depth':
        if value is None:
            return True
        if not isinstance(value, int):
            return False
        if value < 1:
            return False
        return True
    raise ValueError("Unknown option name {}".format(repr(name)))



from torch.utils.data import DataLoader

class BaseDatasetManager:
    """
    Wrapper for DataLoaders. Data attributes:
    - train: DataLoader object for training set
    - test: DataLoader object for test set
    - data_shape: shape of each data point (channels, height, width)
    - img_size: spatial dimensions of each data point (height, width)
    - color_ch: number of color channels
    """

    def __init__(self, cfg, cuda, **kwargs):

        # Define training and test set
        tr_set, ts_set = self.make_datasets(cfg, **kwargs)

        # Dataloaders
        self.train, self.test = self.make_dataloaders(tr_set, ts_set, cfg,
                                                      cuda, **kwargs)

        self.data_shape = self.train.dataset[0][0].size()
        self.img_size = self.data_shape[1:]
        self.color_ch = self.data_shape[0]

    @classmethod
    def make_datasets(cls, cfg, **kwargs):
        """
        Return training and test sets as PyTorch Datasets.
        """
        raise NotImplementedError

    @classmethod
    def make_dataloaders(cls, train, test, cfg, cuda, **kwargs):
        """
        Return dataloader for training and test sets as Pytorch DataLoaders.
        Default dataloaders are provided here; override for custom dataloaders.

        :param train: (Dataset) training set
        :param test: (Dataset) test set
        :param cfg:
        :param cuda:
        :return:
        """

        # Default arguments for dataloaders
        nw = getattr(kwargs, 'num_workers', 0)
        pm = getattr(kwargs, 'pin_memory', False)
        dl_kwargs = {'num_workers': nw, 'pin_memory': pm} if cuda else {}

        dl_train = DataLoader(
            train,
            batch_size=cfg.batch_size,
            shuffle=True,
            drop_last=True,
            **dl_kwargs
        )
        dl_test = DataLoader(
            test,
            batch_size=cfg.test_batch_size,
            shuffle=False,
            **dl_kwargs
        )
        return dl_train, dl_test
